import aiohttp
import asyncio
import logging
import json

from typing import Callable

#< WSMessage(type=<WSMsgType.TEXT: 1>, data='{"serial":"F0AD4E03A9FE","door":1,"door_status":4,"fault":0}', extra='')DEBUG:waiting for event message
# WSMessage(type=<WSMsgType.TEXT: 1>, data='{"serial":"F0AD4E03A9FE","door":1,"door_status":4}', extra='')
#DEBUG:waiting for event message

LOGGER = logging.getLogger(__name__)

WSURI = "wss://event-caster.st1.gdocntl.net/updates"
#WSURI = "wss://app.apps.st1.gdocntl.net/monitor"

class EventSocket:
    def __init__(self, access_token, msg_listener: Callable[[str], None]):
        self._access_token = access_token
        self._msg_listener = msg_listener
        self._running = False
        self._websocket: aiohttp.ClientWebSocketResponse = None
        self._run_future = None
        self._timeout = aiohttp.ClientTimeout(total=None)
       

    async def _run(self):
        if not self._running:
            return
        LOGGER.debug("Started the web socket")
        headers = {
            "Authorization": f'Bearer {self._access_token}'}
        async with aiohttp.ClientSession(timeout=self._timeout, headers=headers) as session:
            async with session.ws_connect(
                WSURI,  heartbeat=20
            ) as ws:
                self._websocket = ws
                LOGGER.debug("Opened the web socket")
                while not ws.closed:
                    LOGGER.debug("waiting for message")
                    msg = await ws.receive() 
                    if not msg:
                        continue
                    LOGGER.debug(f"event message received< {json.loads(msg.data)}")
                    if msg.type == aiohttp.WSMsgType.ERROR:
                        LOGGER.error("Socket message error")
                        break
                    if msg.type in [
                        aiohttp.WSMsgType.CLOSE,
                        aiohttp.WSMsgType.CLOSING,
                        aiohttp.WSMsgType.CLOSED,
                    ]:
                        LOGGER.debug(
                            f"Stopping receiving. Message type: {str(msg.type)}"
                        )
                        break
                    if msg.type != aiohttp.WSMsgType.TEXT:
                        LOGGER.error(f"Socket message type is invalid: {str(msg.type)}")
                        continue
                    
                    await self._msg_listener(msg.data)

        self._websocket = None

        if self._running:
            # Just keep reconnecting - The AladdinConect app just reconnects forever.
            LOGGER.info("Reconnecting...")
            self._run_future = asyncio.get_event_loop().create_task(self._run())

    async def start(self):
        self._running = True
        self._run_future = asyncio.get_event_loop().create_task(self._run())

    async def stop(self):
        self._running = False
        if not self._websocket:
            return
        await self._websocket.close()
        self._websocket = None
        await self._run_future


import logging
import json
from typing import Callable
from AIOAladdinConnect.session_manager import SessionManager
from AIOAladdinConnect.eventsocket import EventSocket

class AladdinConnectClient:
    CONFIGURATION_ENDPOINT = "/configuration"

    DOOR_STATUS_OPEN = 'open'
    DOOR_STATUS_CLOSED = 'closed'
    DOOR_STATUS_OPENING = 'opening'
    DOOR_STATUS_CLOSING = 'closing'
    DOOR_STATUS_UNKNOWN = 'unknown'

    DOOR_COMMAND_CLOSE = "CloseDoor"
    DOOR_COMMAND_OPEN = "OpenDoor"

    DOOR_STATUS = {
        0: DOOR_STATUS_UNKNOWN,  # Unknown
        1: DOOR_STATUS_OPEN,  # open
        2: DOOR_STATUS_OPENING,  # opening
        3: DOOR_STATUS_UNKNOWN,  # Timeout Opening
        4: DOOR_STATUS_CLOSED,  # closed
        5: DOOR_STATUS_CLOSING,  # closing
        6: DOOR_STATUS_UNKNOWN,  # Timeout Closing
        7: DOOR_STATUS_UNKNOWN  # Not Configured
    }

    REQUEST_DOOR_STATUS_COMMAND = {
        DOOR_STATUS_CLOSED: DOOR_COMMAND_CLOSE,
        DOOR_STATUS_OPEN: DOOR_COMMAND_OPEN
    }

    STATUS_CONNECTED = 'Connected'
    STATUS_NOT_CONFIGURED = 'NotConfigured'

    DOOR_LINK_STATUS = {
        0: 'Unknown',
        1: STATUS_NOT_CONFIGURED,
        2: 'Paired',
        3: STATUS_CONNECTED
    }

    CONTROLLER_STATUS = {
        0: 'Offline',
        1: STATUS_CONNECTED
    }

    _LOGGER = logging.getLogger(__name__)

    def __init__(self, email:str, password:str,attr_changed:Callable):
        self._session = SessionManager(email, password)
        self._eventsocket = None
        self._doors = {'device_id':0}
        self._attr_changed = attr_changed
    
    def register_callback(self,update_callback:Callable):
        self._attr_changed = update_callback
        self._LOGGER.debug("Registered callback")
 
    async def login(self):
        self._LOGGER.debug("Logging in")
        # if there is an error, trying to log back needs to stop the eventsocket
        if self._eventsocket:
            await self._eventsocket.stop()

        status = await self._session.login()
        if status:
            self._LOGGER.debug("Logged in")

            await self.get_doors()
            self._LOGGER.debug("Got initial door status")
            
            self._eventsocket = EventSocket(self._session.auth_token(),self._call_back)
            await self._eventsocket.start()
            self._LOGGER.debug("Started Socket")
            
        return status
    
    async def close(self):
        await self._session.close()
        await self._eventsocket.stop()
        return True

    async def get_doors(self):
        """Get all doors status and store values 
            This function should be called intermittently to update all door information"""
        devices = await self._get_devices()

        doors = []

        if devices:
            for device in devices:
                doors += device['doors']
        self._doors = doors
        return doors

    async def _get_devices(self):
        """Get list of devices, i.e., Aladdin Door Controllers"""

        try:

            response = await self._session.get(self.CONFIGURATION_ENDPOINT)
            devices = []
            for device in response["devices"]:
                doors = []
                for door in device["doors"]:
                    doors.append({
                        'device_id': device["id"],
                        'door_number': door["door_index"],
                        'name': door["name"],
                        'status': self.DOOR_STATUS[door["status"]],
                        'link_status': self.DOOR_LINK_STATUS[door["link_status"]],
                        'battery_level': door["battery_level"],
                        'rssi': device['rssi']
                    })
                devices.append({
                    'device_id': device["id"],
                    'doors': doors
                })
            return devices
        except ValueError as ex:
            self._LOGGER.error("Aladdin Connect - Unable to retrieve configuration %s", ex)
            return None

    async def close_door(self, device_id, door_number):
        """Command to close the door"""
        await self._set_door_status(device_id, door_number, self.DOOR_STATUS_CLOSED)

    async def open_door(self, device_id, door_number):
        """Command to open the door"""
        await self._set_door_status(device_id, door_number, self.DOOR_STATUS_OPEN)

    async def _set_door_status(self, device_id, door_number, requested_door_status):
        """Set door state"""
        payload = {"command_key": self.REQUEST_DOOR_STATUS_COMMAND[requested_door_status]}

        try:
            await self._session.call_rpc(f"/devices/{device_id}/door/{door_number}/command", payload)
        except ValueError as ex:
            # Ignore "Door is already open/closed" errors to maintain backwards compatibility
            error = str(ex.args[0])
            should_ignore = error.endswith(
                f'{{"code":400,"error":"Door is already {requested_door_status}"}}')
            if not should_ignore:
                self._LOGGER.error("Aladdin Connect - Unable to set door status %s", ex)
                return False

        return True

    async def get_door_status(self, device_id, door_number):
        for door in self._doors:
            if door["device_id"] == device_id and door["door_number"] == door_number:
                return door["status"]

    async def get_door_link_status(self,device_id,door_number):
        for door in self._doors:
            if door["device_id"] == device_id and door["door_number"] == door_number:
                return door["link_status"]

    async def get_battery_status(self,device_id,door_number):
        for door in self._doors:
            if door["device_id"] == device_id and door["door_number"] == door_number:
                return door["battery_level"]

    async def get_rssi_status(self,device_id,door_number):
        for door in self._doors:
            if door["device_id"] == device_id and door["door_number"] == door_number:
                return door["rssi"]

    async def _call_back(self,msg):
        """Call back from AIO HTTP web socket with door status information"""
        self._LOGGER.debug(f"Got the callback {json.loads(msg)}")
        json_msg = json.loads(msg)
        for door in self._doors:
            if json_msg['door'] == door['door_number']:
                door.update({'status': self.DOOR_STATUS[json_msg["door_status"]]})
                self._LOGGER.debug(f"Status Updated {self.DOOR_STATUS[json_msg['door_status']]}")
                if self._attr_changed:
                    await self._attr_changed()

    def auth_token(self):
        return self._session.auth_token()

    def set_auth_token(self,auth_token):
        self._session.set_auth_token(auth_token)


import base64
import logging
from typing import Any
import aiohttp


class SessionManager:
    HEADER_CONTENT_TYPE_URLENCODED = 'application/x-www-form-urlencoded'
    HEADER_USER_AGENT = "okhttp/3.12.1"
    HEADER_BUNDLE_NAME = "com.geniecompany.AladdinConnect"
    HEADER_BUILD_VERSION = "2042"
    HEADER_APP_VERSION = "5.30"

    API_BASE_URL = "https://pxdqkls7aj.execute-api.us-east-1.amazonaws.com/Android"
    RPC_URL = API_BASE_URL

    LOGIN_ENDPOINT = "/oauth/token"
    X_API_KEY = "fkowarQ0dX9Gj1cbB9Xkx1yXZkd6bzVn5x24sECW"
    

    _LOGGER = logging.getLogger(__name__)

    def __init__(self, email, password):
        self._timeout = aiohttp.ClientTimeout(total=30)
        self._session = aiohttp.ClientSession(timeout = self._timeout)
        self._headers = {'Content-Type': self.HEADER_CONTENT_TYPE_URLENCODED,
                                      'AppVersion': self.HEADER_APP_VERSION,
                                      'BundleName': self.HEADER_BUNDLE_NAME,
                                      'User-Agent': self.HEADER_USER_AGENT,
                                      'BuildVersion': self.HEADER_BUILD_VERSION,
                                      'X-Api-Key': self.X_API_KEY}
        self._auth_token = None
        self._user_email = email
        self._password = password
        self._logged_in = False
    

    def auth_token(self):
        return self._auth_token

    def set_auth_token(self,auth_token):
        self._auth_token = auth_token
        self._headers.update({'Authorization': f'Bearer {self._auth_token}'})


    async def login(self):
        self._auth_token = None
        self._logged_in = False
        password_base64 = base64.b64encode(self._password.encode('utf-8')).decode('utf-8')
        payload = {"grant_type": "password",
                    "client_id": "1000",
                    "brand": "ALADDIN",
                    "username": self._user_email,
                    "password": password_base64,
                    "platform": "platform",
                    "model": "Google Pixel 6",
                    "app_version": "5.30",
                    "build_number": "2042",
                    "os_version": "12.0.0"}


        url = self.API_BASE_URL + self.LOGIN_ENDPOINT
        try:
            response = await self._session.post(url ,data=payload,headers=self._headers)

            if response.content_type == "application/json":
                response_json = await response.json()

            if response_json and "access_token" in response_json:
                self._logged_in = True
                self._auth_token = response_json["access_token"]
                self._headers.update({'Authorization': f'Bearer {self._auth_token}'})
                return True
        except ValueError as ex:
            self._LOGGER.error("Aladdin Connect - Unable to login %s", ex)

        return False

    async def close(self):
        if self._session:
            await self._session.close()

    async def get(self,endpoint:str):
        url = self.API_BASE_URL + endpoint
        self._headers.update({'Content-Type': 'application/x-www-form-urlencoded'})
        try:
            response = await self._session.get(url ,headers=self._headers)
            if response.content_type == "application/json":
                return await response.json()

        except ValueError as ex:
            self._LOGGER.error("Aladdin Connect - Unable to get doors %s", ex)

        return None        

    async def call_rpc(self,api, payload=None):
        self._headers.update({'Content-Type': 'application/json'})
        url = self.API_BASE_URL + api
        try:
            response = await self._session.post(url,json=payload,headers=self._headers)
        
        except ValueError as ex:
            self._LOGGER.error("Aladdin Connect - Unable to operate doors %s", ex)
        
        if response.status not in (200, 204):
            msg = f"Aladdin API call ({url}) failed: {response.status}, {await response.read()}"
            raise ValueError(msg)
        
        if response.content_type == "application/json":
            return await response.json()
        
        return None

    async def call_status(self,api, payload=None):
        self._headers.update({'Content-Type': 'application/json'})
        url = self.API_BASE_URL + api
        try:
            response = await self._session.get(url,headers=self._headers)
        
        except ValueError as ex:
            self._LOGGER.error("Aladdin Connect - Unable to listen to doors %s", ex)

        if response.status not in (200, 204):
            msg = f"Aladdin API call ({url}) failed: {response.status}, {response.text}"
            raise ValueError(msg)
        
        if response.content_type == "application/json":
            return await response.json()
        
        return None

